SuperStrict
Rem
bbdoc: Artemis.AVLTree
about:
A AVL tree implementation for the object type.

Thanks to this tutorial: http://cmcrossroads.com/bradapp/ftp/src/libs/C++/AvlTrees.html
EndRem
Module Artemis.AVLTree

Public

Type TAVLTree
	
	Const _LEFT :Int = 0
	Const _RIGHT:Int = 1
	
	Const _HEIGHT_NOCHANGE:Int = 0
	Const _HEIGHT_CHANGE  :Int = 1
	
	Const _HEAVY_LEFT :Int = -1
	Const _BALANCED   :Int =  0
	Const _HEAVY_RIGHT:Int =  1
	
	Const _COMPARE_MIN:Int = -1
	Const _COMPARE_EQ :Int =  0
	Const _COMPARE_MAX:Int =  1

	Field _root:TAVLNode = Null
	
	Method Create:TAVLTree()
		Return Self
	EndMethod
	
	Rem
	bbdoc: Check whether the tree is empty
	return: 1 if the tree is empty - otherwise 0.
	EndRem
	Method isEmpty:Int()
		Return Null = Self._root
	EndMethod
	
	Rem
	bbdoc: Find an object
	returns: True if object exists in tree - otherwise False.
	EndRem
	Method find:Int(value:Object)
	EndMethod
	
	Rem
	bbdoc: Insert an object
	EndRem
	Method insert:Int(value:Object, height_change:Int Var)
		Return Self._insert(value, Self._root, height_change)
	EndMethod
	
	Method _insert:Int(value:Object, node:TAVLNode Var, height_change:Int Var)
		If Null = node Then
			node = New TAVLNode.Create(value)
			height_change = Self._HEIGHT_CHANGE
			Return 1
		EndIf
		
		Local inserted :Int = 1
		Local increase :Int = 0
		Local result   :Int = node._value.compare(value)
		Local direction:Int
		If result < 0 Then
			direction = Self._LEFT
		Else
			direction = Self._RIGHT
		EndIf			
		If Self._COMPARE_EQ <> result Then
			inserted = Self._insert(value, node._children[direction], height_change)
			If Not inserted Then Return inserted
			increase = result * height_change
		Else
			Return 0
		EndIf
		
		node._balance :+ increase
		
		If increase And node._balance Then
			height_change = 1 - Self._rebalance(node)
		Else
			height_change = Self._HEIGHT_NOCHANGE
		EndIf
		
		Return inserted
	EndMethod
	
	Rem
	bbdoc: Delete an object
	EndRem
	Method remove(value:Object)
	EndMethod
	
	Method _rotateOnce:Int(root:TAVLNode Var, direction:Int)
		Local height_change  :Int
		Local other_direction:Int      = 1 - direction
		Local old_root       :TAVLNode = root
		
		' Check if height will change
		If 0 = root._children[other_direction]._balance Then
			height_change = Self._HEIGHT_NOCHANGE
		Else
			height_change = Self._HEIGHT_CHANGE
		EndIf
		
		' Rotate
		root                                = root._children[other_direction]
		old_root._children[other_direction] = root._children[direction]
		root    ._children[direction]       = old_root
		
		' Update balances
		If Self._LEFT = direction Then
			root._balance :- 1
		Else
			root._balance :+ 1
		EndIf
		old_root._balance = -root._balance
		
		Return height_change
	EndMethod
	
	Method _rotateTwice:Int(root:TAVLNode Var, direction:Int)
		Local other_direction    :Int      = 1 - direction
		Local old_root           :TAVLNode = root
		Local old_other_direction:TAVLNode = root._children[other_direction]
		
		' Rotate
		root = old_root._children[other_direction]._children[direction]
		old_root._children[other_direction] = root._children[direction]
		root._children[direction] = old_root
		old_other_direction._children[direction] = root._children[other_direction]
		root._children[other_direction] = old_other_direction
		
		' Update balances
		root._children[Self._LEFT ]._balance = -Max(root._balance, 0)
		root._children[Self._RIGHT]._balance = -Min(root._balance, 0)
		root._balance = 0
		
		Return Self._HEIGHT_CHANGE
	EndMethod
	
	Method _rebalance:Int(node:TAVLNode Var)
		Local height_change:Int = Self._HEIGHT_NOCHANGE
		Local change_root  :Int = False
		
		If node = Self._root Then change_root = True
		
		If node._balance < Self._HEAVY_LEFT Then
			If node._children[Self._LEFT] And Self._HEAVY_RIGHT = node._children[Self._LEFT]._balance Then
				' RL
				height_change = Self._rotateTwice(node, Self._RIGHT)
			Else
				' RR
				height_change = Self._rotateOnce(node, Self._RIGHT)
			EndIf
		ElseIf node._balance > Self._HEAVY_RIGHT
			If node._children[Self._RIGHT] And Self._HEAVY_LEFT = node._children[Self._RIGHT]._balance Then
				' LR
				height_change = Self._rotateTwice(node, Self._LEFT)
			Else
				' LL
				height_change = Self._rotateOnce(node, Self._LEFT)
			EndIf
		EndIf
		
		If change_root Then Self._root = node
		
		Return height_change
	EndMethod
	
	Method _compare:Int(node:TAVLNode, value:Object, compare:Int)
		Select compare
			Case Self._COMPARE_EQ
				Return node._value.compare(value)
			Case Self._COMPARE_MIN
				If Null = node._children[Self._LEFT] Then
					Return Self._COMPARE_EQ
				Else
					Return Self._COMPARE_MIN
				EndIf
			Case Self._COMPARE_MAX
				If Null = node._children[Self._RIGHT] Then
					Return Self._COMPARE_EQ
				Else
					Return Self._COMPARE_MAX
				EndIf
		EndSelect
	EndMethod
	
	Method dump:String()
		Return Self._dump(Self._root)
	EndMethod
	
	Method _dump:String(node:TAVLNode, pre:String="")
		If Null = node Then Return ""
		
'		Local new_pre:String = pre + "->" + node._value.toString()
'		Local result:String = new_pre + "~n"
		
'		result :+ Self._dump(node._children[Self._LEFT ], new_pre)
'		result :+ Self._dump(node._children[Self._RIGHT], new_pre)
		
		If pre <> "" Then pre :+ " -> "
		
		Local result:String = pre + node._value.toString() + ";~n"
		
		result :+ Self._dump(node._children[Self._LEFT ], node._value.toString())
		result :+ Self._dump(node._children[Self._RIGHT], node._value.toString())
		
		Return result
	EndMethod
	
EndType

Type TAVLNode
	
	Field _value   :Object      = Null
	Field _balance :Int         = 0
	Field _children:TAVLNode[2]
	
	Method Create:TAVLNode(value:Object)
		Self._value = value
		Return Self
	EndMethod
	
EndType

Rem
bbdoc: A AVL tree implementation
'EndRem
Type TAVLTree
	
	Field _root:TAVLNode = Null
	
	Method Create:TAVLTree()
		Return Self
	EndMethod
	
	Rem
	bbdoc: Check whether the tree is empty
	return: 1 if the tree is empty - otherwise 0.
'	EndRem
	Method isEmpty:Int()
		Return Null = Self._root
	EndMethod
	
	Rem
	bbdoc: Find an object
	returns: True if object exists in tree - otherwise False.
'	EndRem
	Method find:Int(value:Object)
		If Null = Self._root Then Return 0
		
		Local node:TAVLNode = Self._root
		
		Repeat
			Local compare:Int = node.compareValue(value)
			If compare = 0 Then Return 1
			If compare < 0 Then
				node = node.getRight()
			Else
				node = node.getLeft()
			EndIf
		Until Null = node
		
		Return 0
	EndMethod
	
	Rem
	bbdoc: Insert an object
'	EndRem
	Method insert(value:Object)
		If Null = Self._root Then
			Self._root = New TAVLNode.Create(value)
			Return
		EndIf
		
		Local new_node:TAVLNode = New TAVLNode.Create(value)
		
		Local node     :TAVLNode = Null
		Local next_node:TAVLNode = Self._root
		
		Repeat
			node = next_node
			Local compare:Int = node.compareValue(value)
			If compare = 0 Then Return 1
			If compare < 0 Then
				position = 1
				next_node = node.getRight()
				If Null = next_node Then node.setRight(new_node)
			Else
				next_node = node.getLeft()
				If Null = next_node Then node.setLeft(new_node)
			EndIf
		Until Null = next_node
	EndMethod
	
	Rem
	bbdoc: Delete an object
'	EndRem
	Method remove(value:Object)
		
	EndMethod
	
EndType

Private

Type TAVLNode
	
	Const UNBALANCED_LEFT :Int = -2
	Const MORE_LEFT       :Int = -1
	Const BALANCED        :Int =  0
	Const MORE_RIGHT      :Int =  1
	Const UNBALANCED_RIGHT:Int =  2
	
	Field _value  :Object = Null
	Field _balance:Int    = TAVLNode.BALANCED
	
	Field _left  :TAVLNode = Null
	Field _right :TAVLNode = Null
	Field _parent:TAVLNode = Null
	
	Method Create:TAVLNode(value:Object)
		Self._value = value
		Return Self
	EndMethod
	
	Method getLeft:TAVLNode()
		Return Self._left
	EndMethod
	
	Method getRight:TAVLNode()
		Return Self._right
	EndMethod
	
	Method getParent:TAVLNode()
		Return Self._parent
	EndMethod
	
	Method getValue:Object()
		Return Self._value
	EndMethod
	
	Method getBalance:Int()
		Return Self._balance
	EndMethod
	
	Method setLeft(node:TAVLNode)
		Self._left = node
		If Self._left Then Self._left._parent = Self
	EndMethod
	
	Method setRight(node:TAVLNode)
		Self._right = node
		If Self._right Then Self._right._parent = Self
	EndMethod
	
	Method setParent(node:TAVLNode)
		Self._parent = node
	EndMethod
	
	Method replaceChild(child:TAVLNode, with:TAVLNode)
		If Self._left  = child Then Self._left  = with
		If Self._right = child Then Self._right = with
	EndMethod
	
	Method rotateRight()
		Local parent_parent:TAVLNode = Self._parent._parent
		
		Self._parent.setLeft(Self._right)
		Self.setRight(Self._parent)
		If parent_parent Then
			parent_parent.replaceChild(Self._right, Self)
		EndIf
		Self._parent = parent_parent
	EndMethod
	
	Method rotateLeft()
		Local parent_parent:TAVLNode = Self._parent._parent
		
		Self._parent.setRight(Self._left)
		Self.setLeft(Self._parent)
		If parent_parent Then
			parent_parent.replaceChild(Self._left, Self)
		EndIf
		Self._parent = parent_parent
	EndMethod
	
	Method compareValue:Int(value:Object)
		Return Self._value.compare(value)
	EndMethod
	
	Method compare:Int(with:Object)
		Local node:TAVLNode = TAVLNode(with)
		If Not node Then Return Null
		Return Self._value.compare(node._value)
	EndMethod
	
EndType